#include "mod_io2.h"

// Disable LOG_DEBUG locally
#undef LOD_DEBUG_ENABLE
#include "log_debug.h"

#include "uptime.h"
#include "adc.h"
#include "flash.h"

void MOD_IO2::Setup() {
	i2c_address = FLASH_OptionByteGet(OB_Data0);
	Begin();
}

void MOD_IO2::Setup(uint8_t addr) {
	i2c_address = addr;
	Begin();
}

void MOD_IO2::Begin() {
	// Validate I2C address
	if (i2c_address > 0x7F || i2c_address == 0x00) {
		i2c_address = MOD_IO2_ADDRESS;
	}

	// Enable needed periphery
	RCC_APB2PeriphClockCmd(
		RCC_APB2Periph_GPIOA | 
		RCC_APB2Periph_GPIOC | 
		RCC_APB2Periph_GPIOD |
		RCC_APB2Periph_AFIO  | 
		RCC_APB2Periph_ADC1  | 
		RCC_APB2Periph_TIM1, 
		ENABLE
	);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);

	// GPIO configuration
	// GPIOs
	for (uint8_t gpio=0; gpio < MOD_IO2_GPIO_COUNT; gpio++) {
		GPIOConfig(gpio, GPIO_Mode_IN_FLOATING, true);
	}
	// Relays
	for (uint8_t relay=0; relay < MOD_IO2_RELAY_COUNT; relay++) {
		RelayConfig(relay);
	}
	// PGM1
	GPIO_InitTypeDef cfg = {0};
	cfg.GPIO_Mode = GPIO_Mode_IPU;
	cfg.GPIO_Pin = PGM1_JUMPER.pin;
	cfg.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(PGM1_JUMPER.port, &cfg);

	// I2C configuration
	I2C_Slave_Setup(i2c_address);

	I2C_Slave_SetOnRegister(onRegisterSet);
	I2C_Slave_SetOnRead(onRegRead);
	I2C_Slave_SetOnWrite(onRegWrite);

	LOG_DEBUG("Initialized." EOL);
}

void MOD_IO2::onRegisterSet(uint8_t reg) {
	i2c_reg = NULL;
	for (uint8_t r = 0; r < MOD_IO2_REG_COUNT; r++) {
		if (REG_MAP[r].reg == reg) {
			i2c_reg = &REG_MAP[r];
			return;
		}
	}
}

uint8_t MOD_IO2::onRegRead(uint8_t reg, uint8_t offset) {
	if (i2c_reg == NULL || i2c_reg->onRead == NULL) {
		return 0x00;
	}

	return i2c_reg->onRead(reg, offset);
}

void MOD_IO2::onRegWrite(uint8_t reg, uint8_t offset, uint8_t data) {
	if (i2c_reg == NULL || i2c_reg->onWrite == NULL) {
		return;
	}
	
	i2c_reg->onWrite(reg, offset, data);
}

uint8_t MOD_IO2::onGetBoardID(uint8_t reg, uint8_t offset) {
	LOG_DEBUG("Board ID 0x%02X" EOL, MOD_IO2_BOARD_ID);
	return MOD_IO2_BOARD_ID;
}

uint8_t MOD_IO2::onGetVersion(uint8_t reg, uint8_t offset) {
	uint8_t version = (MOD_IO2_VER_MAJOR << 4) | MOD_IO2_VER_MINOR;
	LOG_DEBUG("Version 0x%02X" EOL, version);
	return version;
}

void MOD_IO2::onSetDir(uint8_t reg, uint8_t offset, uint8_t data) {
	for (uint8_t gpio = 0; gpio < MOD_IO2_GPIO_COUNT; gpio++) {
		uint8_t mask = (1 << gpio);
		GPIOConfig(
			gpio, 
			((data & mask) != 0 ?
				// Input
				((gpio_pullup & mask) != 0 ? GPIO_Mode_IPU : GPIO_Mode_IN_FLOATING)
				:
				// Outout
				GPIO_Mode_Out_PP
			)
		);
	}
	gpio_direction = data;
	LOG_DEBUG("SetDir " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(gpio_direction));
}

uint8_t MOD_IO2::onGetDir(uint8_t reg, uint8_t offset) {
	LOG_DEBUG("GetDir " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(gpio_direction));
	return gpio_direction;
}

void MOD_IO2::onSetPullUp(uint8_t reg, uint8_t offset, uint8_t data) {
	// clear pullup bit for outputs
	gpio_pullup = (data & gpio_direction);

	for (uint8_t gpio = 0; gpio < MOD_IO2_GPIO_COUNT; gpio++) {
		uint8_t mask = (1 << gpio);
		if ((gpio_direction & mask) == 0) {
			// Skip outputs
			continue;
		}

		GPIOConfig(gpio, (gpio_pullup & mask) != 0 ? GPIO_Mode_IPU : GPIO_Mode_IN_FLOATING);
	}
	LOG_DEBUG("SetPullup " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(gpio_pullup));
}

uint8_t MOD_IO2::onGetPullUp(uint8_t reg, uint8_t offset) {
	LOG_DEBUG("GetPullup " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(gpio_pullup));
	return gpio_pullup;
}

void MOD_IO2::onSetLevel(uint8_t reg, uint8_t offset, uint8_t data) {
	for (uint8_t gpio = 0; gpio < MOD_IO2_GPIO_COUNT; gpio++) {
		uint8_t mask = (1 << gpio);
		// Check if GPIO is output
		if ((gpio_direction & mask) != 0) {
			continue;
		}
		GPIOSet(gpio, (BitAction)((data & mask) != 0));
	}
	LOG_DEBUG("SetGPIO " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(gpio_state));
}

uint8_t MOD_IO2::onGetLevel(uint8_t reg, uint8_t offset) {
	for (uint8_t gpio = 0; gpio < MOD_IO2_GPIO_COUNT; gpio++) {
		GPIOGet(gpio);
	}
	LOG_DEBUG("GetGPIO " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(gpio_state));
	return gpio_state;
}

uint8_t MOD_IO2::onAnalogGet(uint8_t reg, uint8_t offset) {
	if (offset == 1) {
		if (adc_error) {
			LOG_DEBUG("ADC 0x%02X n/a" EOL, reg);
		} else {
			#ifdef MOD_IO2_ADC_VOLTAGE
			LOG_DEBUG("ADC 0x%02X %d %d mV" EOL, reg, adc_value, adc_voltage);
			#else
			LOG_DEBUG("ADC 0x%02X %d" EOL, reg, adc_value);
			#endif
		}
		return ((adc_value >> 8) & 0xFF);
	}

	if (offset > 1) {
		return 0x00;
	}

	uint8_t gpio = reg & 0x0F;
	adc_error = false;

	if (GPIO_MAP[gpio].adc == 255) {
		adc_error = true;
	} else {
		GPIOConfig(gpio, GPIO_Mode_AIN);
		adc_error |= (ADC_SUCCESS != ADC_Channel_Init(GPIO_MAP[gpio].port, GPIO_MAP[gpio].pin, MOD_IO2_ADC_TIMEOUT));
		adc_error |= (ADC_SUCCESS != Get_ADC_Val(GPIO_MAP[gpio].adc, MOD_IO2_ADC_TIMEOUT, &adc_value));
	}

	if (adc_error) {
		adc_value = 0;
		adc_voltage = 0;
	}
	#ifdef MOD_IO2_ADC_VOLTAGE
	 else {
		adc_voltage = ADC_Map(adc_value, 0, 1023, 0.0, 3.3);
	}
	#endif

	return (adc_value & 0xFF);
}

void MOD_IO2::onSetRelayState(uint8_t reg, uint8_t offset, uint8_t data) {
	for (uint8_t relay = 0; relay < MOD_IO2_RELAY_COUNT; relay++) {
		RelaySet(relay, (BitAction)((data & (1 << relay)) != 0));
	}
	LOG_DEBUG("SetRelays " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(relay_state));
}

uint8_t MOD_IO2::onGetRelayState(uint8_t reg, uint8_t offset) {
	LOG_DEBUG("GetRelays " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(relay_state));
	return relay_state;
}

void MOD_IO2::onRelayOn(uint8_t reg, uint8_t offset, uint8_t data) {
	for (uint8_t relay = 0; relay < MOD_IO2_RELAY_COUNT; relay++) {
		if (((data & (1 << relay)) != 0)) {
			RelaySet(relay, Bit_SET);
		}
	}
	LOG_DEBUG("RelaysOn " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(relay_state));
}

void MOD_IO2::onRelayOff(uint8_t reg, uint8_t offset, uint8_t data) {
	for (uint8_t relay = 0; relay < MOD_IO2_RELAY_COUNT; relay++) {
		if (((data & (1 << relay)) != 0)) {
			RelaySet(relay, Bit_RESET);
		}
	}
	LOG_DEBUG("RelaysOff " BYTE_TO_BINARY_PATTERN EOL, BYTE_TO_BINARY(relay_state));
}

uint8_t MOD_IO2::onGetRelayOff(uint8_t reg, uint8_t offset) {
	return (~relay_state) & MOD_IO2_RELAY_MASK;
}

void MOD_IO2::onSetPWM(uint8_t reg, uint8_t offset, uint8_t data) {
	uint8_t pwm = reg & 0x0F;
	if (pwm >= MOD_IO2_PWM_COUNT) {
		return;
	}
	
	uint8_t gpio;
	if (pwm == 0) {
		// Disable PWMx where x = data
		if (data == 0 || data >= MOD_IO2_PWM_COUNT) {
			return;
		}
		gpio = PWM_MAP[data].gpio;
		PWMDisable(GPIO_MAP[gpio].tim);
		GPIOConfig(gpio, GPIO_Mode_IN_FLOATING);
		return;
	}

	gpio = PWM_MAP[pwm].gpio;
	GPIOConfig(gpio, GPIO_Mode_AF_PP);
	PWMConfig(GPIO_MAP[gpio].tim, GPIO_MAP[gpio].tch, data);
}

void MOD_IO2::onSetAddress(uint8_t reg, uint8_t offset, uint8_t data) {
	if (GPIO_ReadInputDataBit(PGM1_JUMPER.port, PGM1_JUMPER.pin) == 0) {
		// Set new address only if PGM1_JUMPER is closed
		Setup(data);

		FLASH_Unlock();
		FLASH_OptionByteSet(OB_Data0, i2c_address);
		FLASH_Lock();
	}
}

void MOD_IO2::GPIOConfig(uint8_t gpio, GPIOMode_TypeDef mode, bool initial) {
	// LOG_DEBUG("GPIOConfig(%d)" EOL, gpio);
	if (gpio >= MOD_IO2_GPIO_COUNT) {
		return;
	}

	uint8_t mask = (1 << gpio);

	bool changes_detected = initial || gpio_config[gpio] != mode;

	if (!changes_detected) {
		// LOG_DEBUG("GPIO %d - NO CHANGES" EOL, gpio);
		return;
	}

	// Set registers
	if ((mode & 0x10) == 0) {
		// input
		gpio_direction |= mask;
		if (mode == GPIO_Mode_IPU) {
			gpio_pullup |= mask;
		} else {
			gpio_pullup &= ~mask;
		}
	} else {
		// output
		gpio_direction &= ~mask;
		gpio_pullup &= ~mask;
	}

	GPIO_InitTypeDef cfg = {0};
	cfg.GPIO_Mode = mode;
	cfg.GPIO_Pin = GPIO_MAP[gpio].pin;
	cfg.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIO_MAP[gpio].port, &cfg);

	gpio_config[gpio] = mode;

	LOG_DEBUG(
		"GPIO %d - %s %s" EOL, 
		gpio, 
		((gpio_direction & mask) != 0 ? "In" : "Out"), 
		((gpio_direction & mask) != 0 ?
			(mode == GPIO_Mode_AIN ?
				"ADC"
				:
				((gpio_pullup & mask) != 0 ? "PU" : "PD")
			)
			:
			"PP"
		)
	);
}

void MOD_IO2::GPIOSet(uint8_t gpio, BitAction level) {
	GPIO_WriteBit(GPIO_MAP[gpio].port, GPIO_MAP[gpio].pin, level);
	if (level) {
		gpio_state |= (1 << gpio);
	} else {
		gpio_state &= ~(1 << gpio);
	}
}

uint8_t MOD_IO2::GPIOGet(uint8_t gpio) {
	uint8_t level;

	if (((uint8_t)gpio_config[gpio] & 0x10) == 0) {
		level = GPIO_ReadInputDataBit(GPIO_MAP[gpio].port, GPIO_MAP[gpio].pin);
	} else {
		level = GPIO_ReadOutputDataBit(GPIO_MAP[gpio].port, GPIO_MAP[gpio].pin);
	}

	if (level) {
		gpio_state |= (1 << gpio);
	} else {
		gpio_state &= ~(1 << gpio);
	}

	return level;
}

void MOD_IO2::RelayConfig(uint8_t relay) {
	LOG_DEBUG("RelayConfig(%d)" EOL, relay);
	if (relay >= MOD_IO2_RELAY_COUNT) {
		return;
	}

	GPIO_InitTypeDef cfg = {0};

	cfg.GPIO_Mode = GPIO_Mode_Out_PP;
	cfg.GPIO_Speed = GPIO_Speed_50MHz;
	cfg.GPIO_Pin = RELAY_MAP[relay].pin;

	GPIO_Init(RELAY_MAP[relay].port, &cfg);
}

void MOD_IO2::RelaySet(uint8_t relay, BitAction level) {
	GPIO_WriteBit(RELAY_MAP[relay].port, RELAY_MAP[relay].pin, level);
	if (level) {
		relay_state |= (1 << relay);
	} else {
		relay_state &= ~(1 << relay);
	}
}

void MOD_IO2::PWMDisable(TIM_TypeDef *TIM) {
	TIM_Cmd(TIM, DISABLE);
	TIM_CtrlPWMOutputs(TIM, DISABLE);
}

void MOD_IO2::PWMConfig(TIM_TypeDef *TIM, uint8_t channel, uint16_t pulse) {
	TIM_Cmd(TIM, DISABLE);
	TIM_CtrlPWMOutputs(TIM, DISABLE);

	TIM_TimeBaseInitTypeDef TimerConfig;
 	TimerConfig.TIM_Period = 255;
	TimerConfig.TIM_Prescaler = 608;
	TimerConfig.TIM_CounterMode = TIM_CounterMode_Up;
	TimerConfig.TIM_ClockDivision = TIM_CKD_DIV1;	
	TIM_TimeBaseInit(TIM, &TimerConfig);

	TIM_OCInitTypeDef TIM_OCConfig={0};
	TIM_OCConfig.TIM_Pulse = pulse;
	TIM_OCConfig.TIM_OCMode = TIM_OCMode_PWM1;
	TIM_OCConfig.TIM_OutputState = TIM_OutputState_Enable;
	TIM_OCConfig.TIM_OCPolarity = TIM_OCPolarity_High;

	switch (channel) {
		case 1:
			TIM_OC1Init(TIM, &TIM_OCConfig );
		break;
		case 2:
			TIM_OC2Init(TIM, &TIM_OCConfig );
		break;
		case 3:
			TIM_OC3Init(TIM, &TIM_OCConfig );
		break;
		case 4:
			TIM_OC4Init(TIM, &TIM_OCConfig );
		break;
	}

	TIM_CtrlPWMOutputs(TIM, ENABLE);

	switch (channel) {
		case 1:
			TIM_OC1PreloadConfig(TIM, TIM_OCPreload_Disable);
		break;
		case 2:
			TIM_OC2PreloadConfig(TIM, TIM_OCPreload_Disable);
		break;
		case 3:
			TIM_OC3PreloadConfig(TIM, TIM_OCPreload_Disable);
		break;
		case 4:
			TIM_OC4PreloadConfig(TIM, TIM_OCPreload_Disable);
		break;
	}

	TIM_ARRPreloadConfig(TIM, ENABLE);

	TIM_Cmd(TIM, ENABLE);
}
